import logging
from typing import TYPE_CHECKING, Any

from admin_extra_buttons.decorators import button, link
from adminfilters.value import ValueFilter
from django.contrib.admin import register
from django.core.cache import caches
from django.db.models.functions import Collate
from django.urls import NoReverseMatch
from smart_admin.modeladmin import SmartModelAdmin

from ..admin_sync import SyncMixin
from ..models import OptionSet
from ..utils import render
from .base import ConcurrencyVersionAdmin

if TYPE_CHECKING:
    from admin_extra_buttons.buttons import LinkButton
    from django.db.models import QuerySet
    from django.http import HttpRequest, HttpResponse

logger = logging.getLogger(__name__)

cache = caches["default"]


@register(OptionSet)
class OptionSetAdmin(SyncMixin, ConcurrencyVersionAdmin, SmartModelAdmin):
    list_display = (
        "name",
        "id",
        "separator",
        "comment",
        "pk_col",
    )
    search_fields = ("name_deterministic",)
    list_filter = (("data", ValueFilter.factory(lookup_name="icontains")),)
    save_as = True
    readonly_fields = ("version", "last_update_date")
    object_history_template = "reversion-compare/object_history.html"
    exclude = ("columns",)

    def get_queryset(self, request: "HttpRequest") -> "QuerySet":
        return super().get_queryset(request).annotate(name_deterministic=Collate("name", "und-x-icu"))

    @button()  # type: ignore[arg-type]
    def display_data(self, request: "HttpRequest", pk: str) -> "HttpResponse":
        ctx = self.get_common_context(request, pk, title="Data")
        obj: OptionSet = ctx["original"]
        data = [line.split(obj.separator) for line in obj.data.split("\r\n")]
        ctx["data"] = data
        return render(request, "admin/core/optionset/table.html", ctx)

    @link(change_form=True, change_list=False, html_attrs={"target": "_new"})  # type: ignore[arg-type]
    def view_json(self, button: "LinkButton") -> None:
        original = button.context["original"]
        if original:
            try:
                button.href = original.get_api_url()
            except NoReverseMatch:
                button.href = "#"
                button.label = "Error reversing url"

    def change_view(
        self, request: "HttpRequest", object_id: str, form_url: str = "", extra_context: "dict[str,Any]|None" = None
    ) -> "HttpResponse":
        if request.method == "POST" and "_saveasnew" in request.POST:
            object_id = None

        return super().change_view(request, object_id)
