import csv
from typing import TYPE_CHECKING

from adminactions.api import delimiters, quotes
from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.forms import BaseFormSet
from django.forms.utils import ErrorList
from django.utils import formats
from django.utils.translation import gettext as _

from .fields.widgets import JavascriptEditor
from .version_media import VersionMedia

if TYPE_CHECKING:
    from .models import CustomFieldType, FormSet


class ValidatorForm(forms.ModelForm):
    code = forms.CharField(widget=JavascriptEditor)


class Select2Widget(forms.Select):
    template_name = "django/forms/widgets/select2.html"


class CustomFieldMixin:
    custom: "CustomFieldType"


class FlexFormBaseForm(forms.Form):
    flex_form = None
    compilation_time_field = None
    indexes = {"1": None, "2": None, "3": None}

    def get_counters(self, data):
        if self.compilation_time_field:
            return data.pop(self.compilation_time_field, {})
        return {}

    def is_valid(self):
        return super().is_valid()

    @property
    def media(self):
        extra = "" if settings.DEBUG else ".min"
        base = super().media
        return (
            VersionMedia(
                js=[
                    "admin/js/vendor/jquery/jquery%s.js" % extra,
                    "admin/js/jquery.init.js",
                    "jquery.compat%s.js" % extra,
                    "smart_validation%s.js" % extra,
                    "smart%s.js" % extra,
                    "smart_field%s.js" % extra,
                ]
            )
            + base
        )

    def get_storage_mapping(self):
        return {name: field.storage for name, field in self.fields.items()}

    def _clean_fields(self):
        super()._clean_fields()
        for name, field in self.fields.items():
            if not field.is_stored():
                del self.cleaned_data[name]

    def full_clean(self):
        return super().full_clean()

    def clean(self):
        cleaned_data = self.cleaned_data
        # if self.is_valid() and self.flex_form and self.flex_form.validator:
        if self.flex_form.validator:
            try:
                self.flex_form.validator.validate(cleaned_data)
            except ValidationError as e:
                raise ValidationError(e) from None
        return cleaned_data


class SmartBaseFormSet(BaseFormSet):
    fs: "FormSet"
    required: bool

    def non_form_errors(self) -> "ErrorList":
        return super().non_form_errors()

    def clean(self):
        if self.fs.validator:
            data = {
                "total_form_count": self.total_form_count(),
                "errors": self._errors,
                "non_form_errors": self._non_form_errors,
                "cleaned_data": getattr(self, "cleaned_data", []),
            }
            try:
                self.fs.validator.validate(data)
            except ValidationError as e:
                raise ValidationError([e]) from None

    @property
    def media(self):
        extra = "" if settings.DEBUG else ".min"
        base = super().media
        return (
            VersionMedia(
                js=[
                    "admin/js/vendor/jquery/jquery%s.js" % extra,
                    "admin/js/jquery.init.js",
                    "jquery.compat%s.js" % extra,
                    "select2/ajax_select%s.js" % extra,
                    "jquery.formset%s.js" % extra,
                    "smart.formset%s.js" % extra,
                ]
            )
            + base
        )


class FormWithDefault(forms.Form):
    defaults: dict[str, str | bool | int]


class DateFormatsForm(FormWithDefault, forms.Form):
    defaults = {
        "date_format": formats.get_format("DATE_FORMAT"),
        "datetime_format": formats.get_format("DATETIME_FORMAT"),
        "time_format": formats.get_format("TIME_FORMAT"),
    }
    datetime_format = forms.CharField(label=_("Datetime format"), required=False)
    date_format = forms.CharField(label=_("Date format"), required=False)
    time_format = forms.CharField(label=_("Time format"), required=False)


class CSVOptionsForm(FormWithDefault, forms.Form):
    defaults = {
        "date_format": "d/m/Y",
        "datetime_format": "N j, Y, P",
        "time_format": "P",
        "doublequote": True,
        "header": False,
        "skipinitialspace": True,
        "quotechar": '"',
        "quoting": csv.QUOTE_MINIMAL,
        "delimiter": ";",
        "escapechar": "\\",
    }

    header = forms.BooleanField(label=_("Header"), required=False)
    doublequote = forms.BooleanField(
        label=_("Doublequote"),
        required=False,
        help_text=_(
            "Controls how instances of quotechar appearing inside a field should "
            "themselves be quoted. When True, the character is doubled. "
            "When False, the escapechar is used as a prefix to the quotechar"
        ),
    )
    skipinitialspace = forms.BooleanField(
        label=_("Skip Initial Space"),
        required=False,
        help_text=_("When True, whitespace immediately following the delimiter is ignored"),
    )
    delimiter = forms.ChoiceField(
        label=_("Delimiter"),
        required=False,
        choices=list(zip(delimiters, delimiters, strict=True)),
        help_text=_("A one-character string used to separate fields"),
    )
    quotechar = forms.ChoiceField(
        label=_("Quotechar"),
        required=False,
        choices=list(zip(quotes, quotes, strict=True)),
        help_text=_(
            "A one-character string used to quote fields containing special characters, "
            "such as the delimiter or quotechar, or which contain new-line characters"
        ),
    )
    quoting = forms.TypedChoiceField(
        coerce=int,
        required=False,
        label=_("Quoting"),
        choices=(
            (csv.QUOTE_ALL, _("All")),
            (csv.QUOTE_MINIMAL, _("Minimal")),
            (csv.QUOTE_NONE, _("None")),
            (csv.QUOTE_NONNUMERIC, _("Non Numeric")),
        ),
        help_text=_("Controls when quotes should be generated by the writer and recognised by the reader"),
    )

    escapechar = forms.ChoiceField(
        label=_("Escapechar"),
        choices=(("", ""), ("\\", "\\")),
        required=False,
        help_text=_(
            "A one-character string used by the writer to escape the delimiter "
            "if quoting is set to QUOTE_NONE and the quotechar if doublequote is False. "
            "On reading, the escapechar removes any special meaning from the following character"
        ),
    )
